define([
// dojo
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/promise/all",
    "dojox/html/entities",

// epi
    "epi/dependency",
    "epi/shell/DestroyableByKey",
    "epi-cms/ApplicationSettings",
    "epi-cms/contentediting/_ContentEditingNotification",
    "epi-cms/contentediting/command/CreateDraft",
    "epi-cms/core/ContentReference",

    // template
    "dojo/text!./templates/ProjectNotification.html",

// resources
    "epi/i18n!epi/cms/nls/episerver.cms.components.project.notifications"
], function(
// dojo
    declare,
    lang,
    all,
    entities,
// epi
    dependency,
    DestroyableByKey,
    ApplicationSettings,
    _ContentEditingNotification,
    CreateDraft,
    ContentReference,

//template
    template,

// resources
    resources
) {

    return declare([_ContentEditingNotification, DestroyableByKey], {
        // summary:
        //      Show notification if the content is part of a project
        // tags:
        //      internal

        // projectService: [readonly] ProjectService
        //      A service for interacting with projects
        projectService: null,

        // _createDraftCommand: [private] Command
        //      The create draft command that will be displayed when the current
        //      content is not part of the selected project.
        _createDraftCommand: null,

        _currentContentLink: null,

        postscript: function() {
            this.inherited(arguments);

            this.projectService = this.projectService || dependency.resolve("epi.cms.ProjectService");

            this.own(
                this.projectService.on("items-added", lang.hitch(this, this._onItemsAddedOrRemoved)),
                this.projectService.on("items-removed", lang.hitch(this, this._onItemsAddedOrRemoved))
            );

            this._createDraftCommand = new CreateDraft({ ignoreContentStatus: true });
        },

        _onExecuteSuccess: function(contentModel) {
            if (!contentModel || !contentModel.contentLink) {
                this._currentContentLink = null;
                return;
            }

            this._createDraftCommand.set("model", contentModel);

            // Watch the model for changes to the content link since will occur when a new draft is created.
            var ownKey = "contentLinkWatch";
            this.destroyByKey(ownKey);
            this.ownByKey(ownKey, contentModel.watch("contentLink", lang.hitch(this, "_onModelChanged", contentModel)));

            // Invoke the model change callback since we have just set the model.
            this._onModelChanged(contentModel);
        },

        _onItemsAddedOrRemoved: function(data) {
            if (!this._currentContentLink) {
                return;
            }

            var hasContentLinkEqualTo = function(currentContentLink) {
                return function(item) {
                    return ContentReference.compareIgnoreVersion(currentContentLink, item.contentLink);
                };
            };

            if (data.items.some(hasContentLinkEqualTo(this._currentContentLink))) {
                this._updateNotification(this._currentContentLink);
            }
        },

        _onModelChanged: function(contentModel) {
            // summary:
            //      Updates the notification when the model changes.
            // tags:
            //      private
            this._currentContentLink = contentModel.contentLink;

            var languageContext = contentModel.languageContext,
                isTranslationNeeded =
                    languageContext &&
                    languageContext.isTranslationNeeded &&
                    ApplicationSettings.currentContentLanguage == languageContext.preferredLanguage;

            this._updateNotification(contentModel.contentLink, contentModel.contentData.isPartOfAnotherProject, isTranslationNeeded);
        },

        _updateNotification: function(contentLink, isPartOfAnotherProject, isTranslationNeeded) {
            var self = this;

            all([this.projectService.getProjectsForContent(contentLink), this.projectService.getCurrentProject()]).then(function(result) {
                var projects = result[0],
                    currentProject = result[1];

                if (projects.length === 0) {
                    self._setNotification(null);
                    return;
                }

                var projectNames = projects.map(function(project) {
                    return entities.encode(project.name);
                }).join(", ");

                //Create the part of project text
                var partOfProjectText = "";
                if(isPartOfAnotherProject) {
                    partOfProjectText = resources.partofanotherproject + ": " + projectNames;
                } else if (currentProject) {
                    partOfProjectText = resources.partofcurrentproject;
                } else {
                    partOfProjectText = resources.partofproject + ": " + projectNames;
                }

                // Replace the text in the template
                var content = lang.replace(template, {
                    partOfProject: partOfProjectText
                });

                var commands =
                    isPartOfAnotherProject && self.projectService.isProjectModeEnabled && !isTranslationNeeded ?
                        [self._createDraftCommand] : [];

                self._setNotification({
                    content: content,
                    commands: commands
                });
            });
        }
    });
});
