﻿define([
// dojo
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/when",
    "dojo/Deferred",
    "dojo/promise/all",
    "dojo/Evented",

// epi
    "epi/dependency",
    "epi/shell/xhr/ErrorHandler",
    "epi-cms/ApplicationSettings",
    "epi-cms/core/ContentReference"
],

function (
// dojo
    declare,
    lang,
    when,
    Deferred,
    all,
    Evented,

// epi
    dependency,
    ErrorHandler,
    ApplicationSettings,
    ContentReference
) {

    return declare([Evented], {
        // summary:
        //      A service for interacting with projects
        // tags:
        //      internal

        // isProjectModeEnabled: [readonly] Boolean
        //      Indicates whether project mode is enabled.
        isProjectModeEnabled: false,

        // projectItemStore: [readonly] Store
        //      A REST store for interacting with project items.
        projectItemStore: null,

        // projectStore: [readonly] Store
        //      A REST store for interacting with projects.
        projectStore: null,

        _currentProjectProfileKey: "epi.current-project-id",

        constructor: function(params) {
            declare.safeMixin(this, params);

            this.projectItemStore = this.projectItemStore || dependency.resolve("epi.storeregistry").get("epi.cms.project.item");
            this.projectStore = this.projectStore || dependency.resolve("epi.storeregistry").get("epi.cms.project");
            this.profile = this.profile || dependency.resolve("epi.shell.Profile");

            this.isProjectModeEnabled = ApplicationSettings.isProjectModeEnabled;
        },

        getProjectItemsForContent: function(contentLinks) {
            // summary:
            //      Get the associated projects items for the given content links
            //
            // contentLinks: String|Array
            //      The content links to get the project items for for
            // tags:
            //      internal

            if(!(contentLinks instanceof Array)){
                contentLinks = [contentLinks];
            }

            return when(this.projectItemStore.query({contentLinks: contentLinks}));
        },

        getProjectsForContent: function(contentLinks) {
            // summary:
            //      Get the associated projects for the given content links
            //
            // contentLinks: String|Array
            //      The content links to get the projects for
            // tags:
            //      internal

            var self = this;

            if(!(contentLinks instanceof Array)){
                contentLinks = [contentLinks];
            }

            return this.getProjectItemsForContent(contentLinks).then(function(projectItems) {
                // If no project items where found return an empty array
                if (projectItems.length === 0) {
                    return new Deferred().resolve([]);
                }

                return self.getProjects(projectItems);
            });
        },

        canAddContent: function(projectId, contentLinks) {
            // summary:
            //      Verifies that the given content links can be added to the project
            //
            // projectId: Number
            //      The project id to check
            // contentLinks: Array
            //      The content links to add
            // returns: Promise
            //      A promise with the result from the server
            // tags:
            //      internal

            return this.projectStore.executeMethod("CanAddContent", projectId, {contentReferences: contentLinks});
        },

        getProjects: function(projectItems) {
             // summary:
            //      Get the projects for the project items
            //
            // projectItems: int[]
            //      The project items to get the projects for
            // tags:
            //      internal

            var projectStore = this.projectStore;
            var projectPromises = projectItems.map(function(projectItem) {
                return projectStore.get(projectItem.projectId);
            });

            return all(projectPromises);
        },

        exists: function(projectId) {
            // summary:
            //      Checks if the given project still exists
            // projectId: Number
            //      The id of the project
            // returns: Promise
            //      True if it still exists otherwise False

            return when(this.projectStore.executeMethod("exists", projectId)).otherwise(ErrorHandler.forXhr);
        },

        addProjectItems: function (projectId, contentReferences) {
            // summary:
            //      Adds the given content references as items of the project.
            //
            // projectId: Number
            //      The target project for adding items to.
            // contentReferences: ContentReference[]
            //      An array of references to content items to add to the project.
            // tags:
            //      public

            // Ensure the content reference we received are valid.
            contentReferences = contentReferences.filter(function (reference) {
                if (ContentReference.isContentReference(reference)) {
                    return true;
                }
                console.warn("epi-cms/project/ProjectService: the given value '" + reference + "' is not a valid content reference.");
            });

            var emit = lang.hitch(this, "emit"),
                store = this.projectItemStore,
                promises = contentReferences.map(function (reference) {
                    return store.add({ contentLink: reference, projectId: projectId });
                }),
                result = all(promises);

            result.then(function(items) {
                    emit("items-added", {
                        items: items
                    });
                })
                .otherwise(ErrorHandler.forXhr);

            return result;
        },

        removeProjectItems: function(projectItems) {
            // summary:
            //      Removes the project items from a project.
            // projectItems: Array
            //      An array of project items to remove from the project
            // tags:
            //      public

            var itemIds = projectItems.map(function(item) { return item.id; }),
                promise = this.projectItemStore.removeRange(itemIds);

            promise
                .then(lang.hitch(this, function() {
                    this.emit("items-removed", {
                        items: projectItems
                    });
                }))
                .otherwise(ErrorHandler.forXhr);

            return promise;
        },

        markAsReadyToPublish: function(projectItems) {
            // summary:
            //      Mark the project items as ready to publish.
            // returns: Promise
            //      A promise that resolves when the items have been marked as ready to publish.
            // tags:
            //      public

            var promise = when(this.projectItemStore.executeMethod("ReadyToPublish", "", projectItems));
            promise.otherwise(ErrorHandler.forXhr);
            return promise;
        },

        isPartOfDelayedPublishedProject: function(contentLink) {
            // summary:
            //      Determines if the content is part of any project that is set to delayed publish
            // contentLink: String
            //      The content link to check
            // returns:
            //      Promise
            // tags:
            //      public

            return this.getProjectsForContent([contentLink])
                .then(function(projects){
                    return projects.some(function(project) {
                        return project.status === "delayedpublished";
                    });
                });
        },

        publishProject: function (projectId, delayPublishUntil) {
            // summary:
            //      Publishes a project with given id right away or delay publishing until given delayPublish date.
            // projectId: Number
            //      Id of the project to publish
            // delayPublishUntil: Date?
            //      If supplied, then the project will be delay published.
            // returns: Promise
            // tags:
            //      public

            var store = this.projectStore;

            if (isNaN(projectId)) {
                throw new Error("projectId is not a number");
            }

            var result = when(store.executeMethod("Publish", projectId, { delayPublishUntil: delayPublishUntil }));
            result.otherwise(ErrorHandler.forXhr);

            // we need to patch the store to update cache and notify the observers.
            return result.then(store.patchCache.bind(store));
        },

        reactivateProject: function(projectId) {
            // summary:
            //      Reactivates a project that has been scheduled for publishing.
            // projectId: Number
            //      The ID of the project to reactivate
            // returns: Promise
            // tags:
            //      public

            var store = this.projectStore;

            if (isNaN(projectId)) {
                throw new Error("The given project ID is not a number.");
            }

            var result = when(store.executeMethod("Reactivate", projectId));
            result.otherwise(ErrorHandler.forXhr);

            // we need to patch the store to update cache and notify the observers.
            return result.then(store.patchCache.bind(store));
        },

        getCurrentProjectId: function() {
            // summary:
            //      Gets the id of the currently selected project from the user profile
            // returns: Promise
            //      Id of the currently selected project as set in the user.
            // tags:
            //      public

            if(!this.isProjectModeEnabled) {
                return new Deferred().resolve(null);
            }
            return when(this.profile.get(this._currentProjectProfileKey));
        },

        getCurrentProject: function() {
            // summary:
            //      Load the current project
            // returns: Promise
            //      The project that was stored in the profile
            // tags:
            //      public

            var projectStore = this.projectStore;

            //Load the selected project from the profile
            return this.getCurrentProjectId().then(function (projectId) {
                if (!projectId) {
                    return null;
                }

                // Need to get the project from the store in order to ensure it still exists.
                // Otherwise return null
                return when(projectStore.get(projectId)).otherwise(function (ex) {console.log(ex); return null; });
            });
        },

        setCurrentProject: function(currentProject) {
            // summary:
            //      Set the current project, it will be stored in the profile. And triggers projectoverview changed
            //  returns: Promise

            this.emit("currentProjectChanged", currentProject);

            return this.profile.set(this._currentProjectProfileKey, currentProject && currentProject.id, { location: "server" });
        }
    });
});
