﻿define("epi/epi", [
    "dojo/_base/lang",
    "dojo/_base/array",
    "dojo/aspect",
    ".",
    "./dependency",
    "./routes",
    "./clientResourcesLoader",
    "./_Module",
    "./ModuleManager",
    "./shell/Bootstrapper",
    "./shell/Stateful",
    "./shell/messaging" // Used for event forwarding between dojo and jQuery. Required for dashboard/mvc gadget compatibility
], function(lang, array, aspect, epi) {

    /* global CSSImportRule */

    // tags:
    //      public

    epi.areEqual = function(first, second) {
        // summary:
        //		are the first and second objects/values deeply
        //		equivalent?

        // Compare undefined always with three equal signs, because undefined==null
        // is true, but undefined===null is false.
        if ((first === undefined) && (second === undefined)) {
            return true;
        }

        if (first === null && second === null) {
            return true;
        }

        // If only one is null, they aren't equal.
        if (first === null || second === null) {
            return false;
        }

        //Are they of the same type?
        if (typeof first !== typeof second) {
            return false;
        }

        if ((first === second) || (first == second) || (typeof first == "number" && typeof second == "number" && isNaN(first) && isNaN(second))) {
            return true;
        }

        //is it an array
        if (lang.isArray(first) && lang.isArray(second)) {
            //Return fals if the length are different
            if (first.length != second.length) {
                return false;
            }

            //Check each value in the array
            for (var x = 0; x < first.length; x++) {
                if (!epi.areEqual(first[x], second[x])) {
                    return false;
                }
            }

            return true;
        }

        //Is it a date
        if ((first instanceof Date) && (second instanceof Date)) {
            //compare number of milliseconds since 1970.
            return first.getTime() === second.getTime();
        }

        //Is it a function
        if((typeof first  === "function") && (typeof second  === "function")){
            return first.toString() === second.toString();
        }

        //Is it an object
        if (((typeof first == "object") && ((typeof second == "object")))) {
            var prop;
            // Make sure ALL THE SAME properties are in both objects!
            for (prop in second) {
                if (!(prop in first)) {
                    return false;
                }
            }

            //Check that the values of the properties are equal
            for (prop in first) {
                if (!epi.areEqual(first[prop], second[prop])) {
                    return false;
                }
            }

            return true;
        }

        return false;
    };

    epi.isEmpty = function(obj) {
        // summary:
        //      Utilities function to check whether an object is empty or not
        //
        // obj: Object
        //
        // tags:
        //      public

        for (var prop in obj) {
            if (obj.hasOwnProperty(prop)) {
                return false;
            }
        }
        return true;
    };

    epi.delegate = function (obj, props, aspectedMethods) {
        // summary:
        //      Create a delegate from the given object and make sure all the "after" aspect advices still work.
        //
        // description:
        //      This function does not ensure the aspect call order between the original object's advices and the delegate object's advice. 
        //      Note that all the aspect advices will receive original arguments and will not be able to change the method result.
        //
        // obj: Object
        //      The original object.
        //
        // props: Object
        //      The override properties
        //
        // aspectedMethods: Array
        //      The methods that might have aspect connect that we want to keep.
        //
        // tags:
        //      public

        // Use dojo delagate to create a delegate object
        var delegate = lang.delegate(obj, props);

        // Flag indicate that the aspect chain is triggered by the aspectBridge advice we added to the delegate object
        var triggerFromDelegate = false;

        // Aspect chain execution helper
        var executeAfterAspect = function (dispatcher) {
            var after = dispatcher.after;
            while (after) {
                after.advice.apply(this, arguments);
                after = after.next;
            }
        };

        // Create aspect bridge handles array, to be able to remove later on.
        delegate._aspectBridges = [];

        // Loop through all the methods
        array.forEach(aspectedMethods, function (propName) {
            if (!obj.hasOwnProperty(propName) || typeof obj[propName] !== "function") {
                return;
            }

            // Add an after advice to the delegate object to trigger the after advices on the original object when the method is called on the delegate object
            delegate._aspectBridges.push(aspect.after(delegate, propName, function () {
                if (triggerFromDelegate) {
                    return;
                }

                var dispatcher = obj[propName];
                executeAfterAspect(dispatcher);
            }, true));

            // Add an after advice to the original object to trigger the after advices on the delegate object when the method is called on the delegate object
            delegate._aspectBridges.push(aspect.after(obj, propName, function () {
                triggerFromDelegate = true;

                var dispatcher = delegate[propName];
                executeAfterAspect(dispatcher);

                triggerFromDelegate = false;
            }, true));
        });

        return delegate;
    };

    epi.removeDelegateAspects = function (delegate) {
        // summary:
        //      Remove the aspect bridges between the delegate object and the original object created when using epi.delegate.
        // delegate: Object
        //      The delegate object.

        delegate._aspectBridges && delegate._aspectBridges.forEach(function (handle) {
            handle.remove();
        });
    };

    epi._countCSSRules = function (asString) {
        // summary:
        //     Counts selectors and rules in all included stylesheets
        // asString:
        //     Boolean
        // tags:
        //     private

        var results = [];

        function countSheet(sheet) {
            var count = 0,
                rule;

            if (sheet && sheet.cssRules) {
                for (var j = 0, l = sheet.cssRules.length; j < l; j++) {
                    rule = sheet.cssRules[j];

                    if(rule instanceof CSSImportRule) {
                        countSheet(rule.styleSheet);
                        continue;
                    }

                    if (rule.selectorText) {
                        count += rule.selectorText.split(',').length;
                    } else if (rule.cssRules) {
                        for (var m = 0, n = rule.cssRules.length; m < n; m++) {
                            if(rule.cssRules[m].selectorText) {
                                count += rule.cssRules[m].selectorText.split(',').length;
                            }
                        }
                    }
                }

                results.push({
                    file: sheet.href ? sheet.href : "inline",
                    rules: sheet.cssRules.length,
                    selectors: count
                });
            }
        }

        if (!document.styleSheets) {
            return;
        }
        for (var i = 0; i < document.styleSheets.length; i++) {
            countSheet(document.styleSheets[i]);
        }

        return asString ? JSON.stringify(results, null, "    ") : results;
    };

    return epi;
});